A link in the form of a wiki link.
Summary
Types
@type t() :: %MDEx.WikiLink{ nodes: [MDEx.Document.md_node()], sourcepos: term(), url: String.t() }
A link in the form of a wiki link.
@type t() :: %MDEx.WikiLink{ nodes: [MDEx.Document.md_node()], sourcepos: term(), url: String.t() }